/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts;

import org.lexgrid.extension.cts.terminology.dto.CodeMap;
import org.lexgrid.extension.cts.terminology.dto.ConceptId;
import org.lexgrid.extension.cts.terminology.dto.MappedConceptCode;
import org.lexgrid.extension.cts.terminology.exceptions.AmbiguousMapRequest;
import org.lexgrid.extension.cts.terminology.exceptions.MapNameSourceMismatch;
import org.lexgrid.extension.cts.terminology.exceptions.MapNameTargetMismatch;
import org.lexgrid.extension.cts.terminology.exceptions.MappingNotAvailable;
import org.lexgrid.extension.cts.terminology.exceptions.UnableToMap;
import org.lexgrid.extension.cts.terminology.exceptions.UnexpectedError;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownCodeSystem;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownConceptCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownMapName;

/**
 * Interface for CTS terminology code mappings. See
 * http://informatics.mayo.edu/LexGrid/downloads/CTS/specification/ctsspec/cts.htm.
 *
 * HL7 messaging software may need to map from local concepts to the standardized codes used in the
 * HL7 messaging environment, or between different standardized code sets. The Code Mapping API
 * provides an interface for performing these mappings.
 *
 * A CodeMap maps some or all of the concept codes from a source CodeSystem to the closest
 * corresponding concept code in the target CodeSystem. A CodeMap is unidirectional - mapping from
 * the source to the target does not imply the ability to map from the target back to the source. A
 * CodeMap may optionally identify the from or to versions of the code systems being translated.
 *
 *
 * Each CodeMap is identified by a unique map_name. It is possible for there to be more than one
 * CodeMap between the same source and target CodeSystems. Each CodeMap has an optional
 * mapDescription attribute that describes the source of the translation, when it was created, how
 * it was done, etc.
 *
 * mapQuality_code signifies the 'quality' of the translation from the source code to the target
 * code, and can be something like "exact", "narrower", "broader" or "partial overlap".
 */
public interface CodeMappingOperations extends IdentifiableService {

    /**
     * Returns the set of code maps that are supported by this particular service. Note that code
     * maps are not symmetric. The fact that a service supports a map from code system A to code
     * system B does not imply that it also supports a map in the reverse direction
     *
     * @return
     * @throws UnexpectedError
     */
    CodeMap[] getSupportedMaps();

    /**
     * Maps (or translates) the supplied code system id and concept code to the corresponding code
     * (if any) in the target system. fromConcept_id must be supplied. Either toCodeSystem_id,
     * map_name or both may be specified. If only map_name is supplied, toCodeSystem_id is
     * determined by the name. If both parameters are supplied, they must be in agreement.
     *
     * @param sourceConceptId
     *            Concept Code, and its Code System identifier, that you want to map, or translate.
     *
     * @param targetCodeSystem
     *            The Code System that you want to map, or translate, to.
     * @param mapName
     * @return the target code system and it's concept code. Also includes the quality code, which
     *         may be something like (this is not an actual value required to return, just a
     *         conceptual example) "isBroaderThan" or "isNarrowerThan".
     * @throws AmbiguousMapRequest
     * @throws UnableToMap
     * @throws MapNameTargetMismatch
     * @throws UnknownMapName
     * @throws MapNameSourceMismatch
     * @throws UnknownConceptCode
     * @throws MappingNotAvailable
     * @throws UnknownCodeSystem
     * @throws UnexpectedError
     */
    MappedConceptCode mapConceptCode(ConceptId sourceConceptId, java.lang.String targetCodeSystem,
            java.lang.String mapName);
}