/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts;

import org.lexgrid.extension.cts.terminology.dto.CodeSystemIdAndVersions;
import org.lexgrid.extension.cts.terminology.dto.CodeSystemInfo;
import org.lexgrid.extension.cts.terminology.dto.ConceptId;
import org.lexgrid.extension.cts.terminology.dto.StringAndLanguage;
import org.lexgrid.extension.cts.terminology.exceptions.CodeSystemNameIdMismatch;
import org.lexgrid.extension.cts.terminology.exceptions.NoApplicableDesignationFound;
import org.lexgrid.extension.cts.terminology.exceptions.TimeoutError;
import org.lexgrid.extension.cts.terminology.exceptions.UnexpectedError;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownCodeSystem;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownConceptCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownLanguageCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownRelationQualifier;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownRelationshipCode;

public interface RuntimeOperations extends IdentifiableService {

/**
     * Determine whether the supplied concept codes are related
     *
     * @param  codeSystemId - code system of the parent and child codes .
     * @param  sourceCode - concept code that occurs as the source of the relationship.
     * @param  targetCode - the concept code that occurs as the target of the relationship
     * @param  relationshipCode -the concept code that identifies the relationship
     * @param  relationQualifiers - an optional list of relationship qualifier codes. If the relationQualifiers list
     *         contains one or more qualifier codes,    only  relationship entries that that have qualifiers that match all
     *         of the qualifiers in the list will be considered.
     * @param  directRelationsOnly - TRUE means test direct relationships only, FALSE means that the transitive closure
     *         of the relationship is to be tested if the relationship is transitive. If the relationship is not transitive, the
     *         result is the same no matter the setting of this flag.
     * @return
     * @throws UnknownRelationshipCode
     * @throws UnknownRelationQualifier
     * @throws UnknownConceptCode
     * @throws UnknownCodeSystem
     * @throws UnexpectedError
     */
boolean areCodesRelated(java.lang.String codeSystemId, java.lang.String sourceCode,
            java.lang.String targetCode, java.lang.String relationshipCode,
            java.lang.String[] relationQualifiers, boolean directRelationsOnly);

    /**
     * Provides a list of all code systems and versions supported by the service.
     *
     * @return a list of all code systems and versions supported by the service.
     * @throws TimeoutError
     * @throws UnexpectedError
     */
    CodeSystemIdAndVersions[] getSupportedCodeSystems();

    /**
     * Provides a list of all code systems and versions supported by the service.
     *  @param timeOut
     *   @param sizeLimit
     * @return a list of all code systems and versions supported by the service.
     * @throws TimeoutError
     * @throws UnexpectedError
     */
    CodeSystemIdAndVersions[] getSupportedCodeSystems(long timeOut, int sizeLimit);

    /**
     * Determine whether concept code is currently valid in the specified code system.
     *
     * @param conceptId - the code system and concept code to validate
     * @param activeConceptsOnly - TRUE means that only active concepts are considered valid. FALSE means that any concept
     *        code known to the code system is ok.
     * @return
     * @throws UnknownCodeSystem
     * @throws UnexpectedError
     */
    boolean isConceptIdValid(ConceptId conceptId, boolean activeConceptsOnly);

    /**
     * Takes a code system identifier (OID) and/or name and returns a detailed description of the
     * code system and elements that are supported by the service.
     *
     * @param codeSystemId - the unique code system identifier, which is usually the ISO OID.
     * @param codeSystemName - the name of the code system.
     * @return
     * @throws UnknownCodeSystem
     * @throws CodeSystemNameIdMismatch
     * @throws UnexpectedError
     */
    CodeSystemInfo lookupCodeSystemInfo(java.lang.String codeSystemId,
            java.lang.String codeSystemName);

    /**
     * Returns the most appropriate designation for the supplied concept identifier in the specified language.
     *
     * @param conceptId - the code system id and concept code
     * @param languageCode- the desired language of the returned designation
     * @return
     * @throws UnknownLanguageCode
     * @throws UnknownConceptCode
     * @throws UnknownCodeSystem
     * @throws NoApplicableDesignationFound
     * @throws UnexpectedError
     */
    StringAndLanguage lookupDesignation(ConceptId conceptId, java.lang.String languageCode);

}