/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.impl;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.LexGrid.LexBIG.DataModel.Collections.CodingSchemeRenderingList;
import org.LexGrid.LexBIG.DataModel.Collections.LocalNameList;
import org.LexGrid.LexBIG.DataModel.Collections.NameAndValueList;
import org.LexGrid.LexBIG.DataModel.Core.ConceptReference;
import org.LexGrid.LexBIG.DataModel.Core.NameAndValue;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.Exceptions.LBInvocationException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet.ActiveOption;
import org.LexGrid.LexBIG.Utility.Iterators.ResolvedConceptReferencesIterator;
import org.LexGrid.commonTypes.Property;
import org.LexGrid.commonTypes.Text;
import org.LexGrid.concepts.Concept;
import org.LexGrid.concepts.Presentation;
import org.apache.log4j.Logger;
import org.lexgrid.extension.cts.BrowserOperations;
import org.lexgrid.extension.cts.lexevs.adapters.CodingSchemeListToIdAndVersionsArrayAdapter;
import org.lexgrid.extension.cts.lexevs.adapters.ConceptIdToConceptReferenceAdapter;
import org.lexgrid.extension.cts.lexevs.adapters.ConceptPropertyAdapter;
import org.lexgrid.extension.cts.lexevs.adapters.ConceptReferenceToCompleteConceptAdapter;
import org.lexgrid.extension.cts.lexevs.adapters.ConceptReferenceToConceptIdAdapter;
import org.lexgrid.extension.cts.lexevs.util.LanguageUtil;
import org.lexgrid.extension.cts.lexevs.util.matchalgorithm.MatchingAlgorithm;
import org.lexgrid.extension.cts.lexevs.util.matchalgorithm.MatchingAlgorithmFactory;
import org.lexgrid.extension.cts.lexevs.util.matchalgorithm.MatchingAlgorithmFactoryImpl;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationGraph;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationMapper;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationRestrictions;
import org.lexgrid.extension.cts.lexevs.util.template.ConceptMapper;
import org.lexgrid.extension.cts.lexevs.util.template.ResolveParams;
import org.lexgrid.extension.cts.lexevs.util.template.Restrictions;
import org.lexgrid.extension.cts.terminology.dto.CodeSystemIdAndVersions;
import org.lexgrid.extension.cts.terminology.dto.CompleteCodedConceptDescription;
import org.lexgrid.extension.cts.terminology.dto.ConceptDesignation;
import org.lexgrid.extension.cts.terminology.dto.ConceptId;
import org.lexgrid.extension.cts.terminology.dto.ConceptProperty;
import org.lexgrid.extension.cts.terminology.dto.ConceptRelationship;
import org.lexgrid.extension.cts.terminology.dto.LexEvsExpansionContext;
import org.lexgrid.extension.cts.terminology.dto.RelatedCode;
import org.lexgrid.extension.cts.terminology.exceptions.BadlyFormedMatchText;
import org.lexgrid.extension.cts.terminology.exceptions.InvalidExpansionContext;
import org.lexgrid.extension.cts.terminology.exceptions.TimeoutError;
import org.lexgrid.extension.cts.terminology.exceptions.UnexpectedError;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownCodeSystem;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownConceptCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownLanguageCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownMatchAlgorithm;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownMimeTypeCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownPropertyCode;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownRelationshipCode;
import org.lexgrid.extension.cts.terminology.impl.exceptions.ExceptionHandler;
import org.lexgrid.extension.cts.terminology.impl.exceptions.ExceptionUtil;

public class LexBigBrowserOperations extends IdentifiableServicesImpl implements BrowserOperations {

    private static Logger log = Logger.getLogger(LexBigBrowserOperations.class);

    private MatchingAlgorithmFactory matchingAlgorithmFactory;

    private String[] supportedMimeTypes;

    private String codeExpansionContainerName;

    private List<ExceptionHandler<Exception>> exceptionHandlers = new ArrayList<ExceptionHandler<Exception>>();

    /**
     * Constructor
     */
    public LexBigBrowserOperations(){
    	matchingAlgorithmFactory = MatchingAlgorithmFactoryImpl.getInstance();
    }
    
/**
     * @return the supportedMimeTypes
     */
public String[] getSupportedMimeTypes() {
        return supportedMimeTypes;
    }

    /**
     * @param supportedMimeTypes
     *            the supportedMimeTypes to set
     */
    public void setSupportedMimeTypes(String[] supportedMimeTypes) {
        this.supportedMimeTypes = supportedMimeTypes;
    }

    /**
     * @return the exceptionHandlers
     */
    public List<ExceptionHandler<Exception>> getExceptionHandlers() {
        return exceptionHandlers;
    }

    /**
     * @param exceptionHandlers
     *            the exceptionHandlers to set
     */
    public void setExceptionHandlers(List<ExceptionHandler<Exception>> exceptionHandlers) {
        this.exceptionHandlers = exceptionHandlers;
    }

    public String getCodeExpansionContainerName() {
        return codeExpansionContainerName;
    }

    public void setCodeExpansionContainerName(String relatedCodeContainerName) {
        this.codeExpansionContainerName = relatedCodeContainerName;
    }

    public MatchingAlgorithmFactory getMatchingAlgorithmFactory() {
        return matchingAlgorithmFactory;
    }

    public void setMatchingAlgorithmFactory(MatchingAlgorithmFactory matchingAlgorithmFactory) {
        this.matchingAlgorithmFactory = matchingAlgorithmFactory;
    }

    /*
     * (non-Javadoc)
     * @see com.gehc.ce.terminology.cts.BrowserOperations#expandCodeExpansionContext(byte[])
     */
    public RelatedCode[] expandCodeExpansionContext(byte[] contextToExpand) {

		RelatedCode[] relatedCodes = {};
		try {
			ByteArrayInputStream in = new ByteArrayInputStream(contextToExpand);
			ObjectInputStream ois = new ObjectInputStream(in);
			LexEvsExpansionContext expContext = (LexEvsExpansionContext) ois
					.readObject();
			ConceptId conceptId = new ConceptId();
			conceptId.setCodeSystemId(expContext.getCodingSchemeId());
			conceptId.setConceptCode(expContext.getConceptCode());

			relatedCodes = this.lookupCodeExpansion(conceptId, expContext
					.getRelationshipCode(), expContext.isSourceToTarget(),
					expContext.isDirectRelationsOnly(), expContext
							.getDesignationLanguagecode(), expContext
							.getDepthLevel());

		} catch (Exception e) {

			if (e instanceof IOException || e instanceof ClassNotFoundException) {
				throw new InvalidExpansionContext();
			} else {
				throw new UnexpectedError(e);
			}

		}
		return relatedCodes;
	}

    /*
	 * (non-Javadoc)
	 * 
	 * @see com.gehc.ce.terminology.cts.BrowserOperations#getSupportedCodeSystems()
	 */
    public CodeSystemIdAndVersions[] getSupportedCodeSystems() {
        // note: no copyright info provided from LexBig
        try {
            CodingSchemeRenderingList supportedCodingSchemes = super.lexEVStemplate.getService()
                    .getSupportedCodingSchemes();
            if (supportedCodingSchemes == null) {
                return new CodeSystemIdAndVersions[0];
            }
            return CodingSchemeListToIdAndVersionsArrayAdapter.adapt(super.lexEVStemplate.getService(), supportedCodingSchemes);

        } catch (Exception e) {
            log.error(e);
            handle(e);
            throw new UnexpectedError(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see com.gehc.ce.terminology.cts.BrowserOperations#getSupportedMatchAlgorithms()
     */
    public String[] getSupportedMatchAlgorithms() {

        List<String> matchAlgorithms = new ArrayList<String>();
        Set<String> keyNames = matchingAlgorithmFactory.getMatchAlgorithmMap().keySet();
        Iterator<String> iterator = keyNames.iterator();
        while (iterator.hasNext()) {
            String supportedMatchAlgorim = iterator.next().toString();
            matchAlgorithms.add(supportedMatchAlgorim);
        }

        return matchAlgorithms.toArray(new String[matchAlgorithms.size()]);
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.BrowserOperations#lookupCodeExpansion(com.gehc.ce.terminology
     * .cts.dto.ConceptId, java.lang.String, boolean, boolean, java.lang.String)
     */
    public RelatedCode[] lookupCodeExpansion(ConceptId expandConceptId, String relationshipCode,
            boolean sourceToTarget, boolean directRelationsOnly, String designationLanguageCode) {

        return lookupCodeExpansion(expandConceptId, relationshipCode, sourceToTarget,
                directRelationsOnly, designationLanguageCode, 0);
    }

    /**
     * Same as public lookupCodeExpansion but includes depthlevel to keep track
     * of the level of expansion.
     * @param expandConceptId
     * @param relationshipCode
     * @param sourceToTarget
     * @param directRelationsOnly
     * @param designationLanguageCode
     * @param depthLevel
     * @return
     */
    private RelatedCode[] lookupCodeExpansion(ConceptId expandConceptId, String relationshipCode,
            boolean sourceToTarget, boolean directRelationsOnly, String designationLanguageCode,
            int depthLevel) {

        depthLevel++;
        verifyCodeSystem(expandConceptId);
        if (expandConceptId.getConceptCode() == null) {
            expandConceptId.setConceptCode("");
        }
        ResolveParams resolveParams = setupCodeExpasionResolveParams(sourceToTarget,
                directRelationsOnly);
        List<RelatedCode> relatedCodeList = new ArrayList<RelatedCode>();
        ConceptReference conceptRef = ConceptIdToConceptReferenceAdapter.adapt(expandConceptId);
        if (conceptRef.getConceptCode() != null && !conceptRef.getConceptCode().equals("")) {
            ResolvedConceptReference queryConcept = queryConcept(expandConceptId);
            if (queryConcept == null) {
                throw new UnknownConceptCode(expandConceptId.getConceptCode());
            }
        }
        AssociationGraph<RelatedCode> graph = super.lexEVStemplate.queryAssociations(conceptRef, createMapper(
                designationLanguageCode, depthLevel), createRestrictions(relationshipCode),
                resolveParams);

        if (sourceToTarget) {
            relatedCodeList.addAll(graph.getWhereFocusIsSource());
        } else {
            relatedCodeList.addAll(graph.getWhereFocusIsTarget());
        }
        List<RelatedCode> expRelatedCodeList = expandRelatedCodes(relatedCodeList,
                directRelationsOnly);
        relatedCodeList.addAll(expRelatedCodeList);

        return relatedCodeList.toArray(new RelatedCode[relatedCodeList.size()]);
    }

    /**
     * Sets values for Resolveparams based on sourceToTarget and directRelationsOnly
     * @param sourceToTarget
     * @param directRelationsOnly
     * @return
     */
    private ResolveParams setupCodeExpasionResolveParams(boolean sourceToTarget,
            boolean directRelationsOnly) {

        ResolveParams resolveParams = new ResolveParams();
        if (!sourceToTarget) {
            resolveParams.setResolveGraphBackward(true);
            resolveParams.setResolveGraphForward(false);
        }
        if (directRelationsOnly) {
            resolveParams.setResolveGraphAssociationDepth(2);
            resolveParams.setResolveGraphCodedEntryDepth(1);
        } else {
            resolveParams.setResolveGraphAssociationDepth(-1);
            resolveParams.setResolveGraphCodedEntryDepth(1);
        }
        return resolveParams;
    }

    /**
     * Iterates through List of relatedCodes to expand those that are able to expand
     * Calls expandCodeExpansionContext
     * @param relatedCodeList
     * @param directRelationsOnly
     * @return
     */
    private List<RelatedCode> expandRelatedCodes(List<RelatedCode> relatedCodeList,
            boolean directRelationsOnly) {
        List<RelatedCode> expRelatedCodeList = new ArrayList<RelatedCode>();
        Iterator<RelatedCode> relatedCodes = relatedCodeList.iterator();
        while (relatedCodes.hasNext()) {
            RelatedCode relatedCode = (RelatedCode)relatedCodes.next();
            if (!directRelationsOnly && relatedCode.isCanExpand()) {
                log.info("Concept Code to be expanded: " + relatedCode.getConceptCode());

                RelatedCode[] expandedCodes = this.expandCodeExpansionContext(relatedCode
                        .getExpansionContext());
                for (RelatedCode expandedCode : expandedCodes) {
                    expRelatedCodeList.add(expandedCode);
                }
            }
        }
        return expRelatedCodeList;
    }

    /**
     * Create an association mapper that will put the association data into a RelatedCode.
     *
     * @param languageCode
     *            TODO
     *
     * @return
     */
    protected AssociationMapper<RelatedCode> createMapper(String languageCode, int depthLevel) {
        return new RelatedCodeAssociationMapper(languageCode, depthLevel);
    }

    /**
     * Restrict the query to relationship code.
     *
     * @param relationshipCode
     * @return
     */
    protected AssociationRestrictions createRestrictions(final String relationshipCode) {
        AssociationRestrictions restrictions = new AssociationRestrictions() {
            public void addRestrictions(CodedNodeGraph nodeGraph) throws LBInvocationException,
                    LBParameterException {
                NameAndValue nameValue = new NameAndValue();
                nameValue.setName(relationshipCode);
                NameAndValueList nameValueList = new NameAndValueList();
                nameValueList.addNameAndValue(nameValue);
                nodeGraph.restrictToAssociations(nameValueList, null);
            }

            public String getContainer() {
                return getCodeExpansionContainerName();
            }

            public void addRestrictions(CodedNodeSet nodeSet) {}
        };
        return restrictions;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.BrowserOperations#lookupCompleteCodedConcept(com.gehc.ce.terminology
     * .cts.dto.ConceptId)
     */
    public CompleteCodedConceptDescription lookupCompleteCodedConcept(ConceptId conceptId) {

        verifyConceptId(conceptId);
        ResolvedConceptReference resolvedConcept = queryConcept(conceptId);

        if (resolvedConcept == null) {
            throw new UnknownConceptCode(conceptId.getConceptCode());
        }

        AssociationGraph<ConceptRelationship> relatedCodesGraph = null;
        try {
            relatedCodesGraph = super.lexEVStemplate.queryAssociations(resolvedConcept,
                    createRelationshipMapper(resolvedConcept), null, createGraphResolveParams());
        } catch (Exception e) {
            log.error(e);
            handle(e);
        }

        return createCompleteConcept(resolvedConcept, relatedCodesGraph);

    }

    /**
     * Given the results from previous LexEVS queries, put together the CTS
     * {@link CompleteCodedConceptDescription}
     *
     * @param resolvedConcept
     * @param relatedCodesGraph
     * @return
     */
    private CompleteCodedConceptDescription createCompleteConcept(
            ResolvedConceptReference resolvedConcept,
            AssociationGraph<ConceptRelationship> relatedCodesGraph) {
        CompleteCodedConceptDescription completeConcept = ConceptReferenceToCompleteConceptAdapter
                .adapt(resolvedConcept);
        if (relatedCodesGraph != null) {
            completeConcept.setSourceFor(toArray(relatedCodesGraph.getWhereFocusIsSource()));
            completeConcept.setTargetOf(toArray(relatedCodesGraph.getWhereFocusIsTarget()));
        }

        return completeConcept;
    }

    /**
     * Query for the concept using the lexEVStemplate and wrap any exceptions in CTS exceptions
     *
     * @param conceptId
     * @return
     */
    private ResolvedConceptReference queryConcept(ConceptId conceptId) {
        try {
            ResolvedConceptReference resolvedConcept = super.lexEVStemplate.queryConcept(conceptId
                    .getConceptCode(), conceptId.getCodeSystemId());
            return resolvedConcept;
        } catch (Exception e) {
            log.error(e);
            handle(e);
        }
        throw new UnexpectedError();
    }

    private ConceptRelationship[] toArray(List<ConceptRelationship> list) {
        return list.toArray(new ConceptRelationship[list.size()]);
    }

    /**
     * @return
     */
    private AssociationMapper<ConceptRelationship> createRelationshipMapper(
            final ConceptReference rootConcept) {
        return new RelationshipMapper(rootConcept);
    }

    /**
     * @return resolve parameters that ensure retrieving both target and source relationships at
     *         once.
     */
    private ResolveParams createGraphResolveParams() {
        ResolveParams parms = new ResolveParams();
        parms.setResolveGraphBackward(true);
        parms.setResolveGraphForward(true);
        return parms;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.BrowserOperations#lookupConceptCodesByDesignation(java.lang.String
     * , java.lang.String, java.lang.String, java.lang.String, boolean)
     */
    public ConceptId[] lookupConceptCodesByDesignation(String codeSystemId, String matchText,
            String matchAlgorithmCode, String languageCode, boolean activeConceptsOnly) {

        try {
            List<ConceptId> conceptIds = super.lexEVStemplate.queryConceptList(codeSystemId,
                    createConceptIdMapper(), createLookupByDesignationRestrictions(matchText,
                            matchAlgorithmCode, languageCode, activeConceptsOnly));
            return conceptIds.toArray(new ConceptId[conceptIds.size()]);
        } catch (Exception e) {
            log.error(e);
            handle(e);
        }
        throw new UnexpectedError();
    }

    /**
     * Handle any exception with the Handlers.
     *
     * @param e
     */
    private void handle(Exception e) {
        ExceptionUtil.handle(e, this.exceptionHandlers);
    }

    /**
     * Get the right {@link MatchingAlgorithm} and create a {@link LookupByDesignationRestrictions}
     * with that and the other given parameters.
     *
     * @return
     */
    private Restrictions createLookupByDesignationRestrictions(String matchText,
            String matchAlgorithmCode, String languageCode, boolean activeConceptsOnly) {

        MatchingAlgorithm matchAlgorithm = getMatchAlgorithm(matchAlgorithmCode);
        return new LookupByDesignationRestrictions(matchText, matchAlgorithm, languageCode,
                activeConceptsOnly);
    }

    /**
     * @return a new {@link ConceptIdMapper}
     */
    private ConceptMapper<ConceptId> createConceptIdMapper() {
        return new ConceptIdMapper();
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.BrowserOperations#lookupConceptCodesByProperty(java.lang.String,
     * java.lang.String, java.lang.String, java.lang.String, boolean, java.lang.String[],
     * java.lang.String[])
     */
    public ConceptId[] lookupConceptCodesByProperty(String codeSystemId, String matchText,
            String matchAlgorithmCode, String languageCode, boolean activeConceptsOnly,
            String[] properties, String[] mimeTypes) {

        try {
            validateMimeTypes(mimeTypes);
            CodedNodeSet codeNodeSet = super.lexEVStemplate.createNodeSet(codeSystemId, null);

            LocalNameList localNameList = new LocalNameList();
            localNameList.setEntry(properties);
            String updatedMatchText = matchText;
            MatchingAlgorithm newMatchAlgorithm = null;
            if (matchAlgorithmCode != null) {
                newMatchAlgorithm = getMatchAlgorithm(matchAlgorithmCode);
                updatedMatchText = newMatchAlgorithm.updateMatchText(matchText);
            }

            codeNodeSet.restrictToMatchingProperties(localNameList, null, updatedMatchText,
                    newMatchAlgorithm.getName(), languageCode);

            if (activeConceptsOnly) {
                codeNodeSet.restrictToStatus(ActiveOption.ACTIVE_ONLY, null);
            }

            ResolvedConceptReferencesIterator resolvedConcepts = codeNodeSet.resolve(null, null,
                    null);
            List<ConceptId> conceptIds = new ArrayList<ConceptId>();
            while (resolvedConcepts.hasNext()) {
                ResolvedConceptReference resolvedConcept = resolvedConcepts.next();
                ConceptId conceptId = ConceptReferenceToConceptIdAdapter.adapt(resolvedConcept);
                conceptIds.add(conceptId);
            }
            return conceptIds.toArray(new ConceptId[conceptIds.size()]);

        } catch (Exception e) {
            log.error(e);
            handle(e);
        }
        throw new UnexpectedError();
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.BrowserOperations#lookupDesignations(com.gehc.ce.terminology.
     * cts.dto.ConceptId, java.lang.String, java.lang.String, java.lang.String)
     */
    public ConceptDesignation[] lookupDesignations(ConceptId conceptCode, String matchText,
            String matchAlgorithmCode, String languageCode) {
        verifyConceptId(conceptCode);
        List<ConceptDesignation> conceptDesignationList = new ArrayList<ConceptDesignation>();
        ResolvedConceptReference resolvedConceptReference = queryConcept(conceptCode);
        if (resolvedConceptReference == null) {
            throw new UnknownConceptCode(conceptCode.getConceptCode());
        }
        Concept c = resolvedConceptReference.getReferencedEntry();
        Presentation[] pArray = c.getPresentation();
        for (Presentation p : pArray) {
            if (doesItMatch(matchAlgorithmCode, p.getValue().getContent(), matchText)
                    && (isValidLanguage(languageCode, p))) {
                ConceptDesignation cd = new ConceptDesignation();
                cd.setDesignation(p.getValue().getContent());
                conceptDesignationList.add(cd);
            }
        }
        return conceptDesignationList
                .toArray(new ConceptDesignation[conceptDesignationList.size()]);
    }

    /**
     * Check if the text matches based on the given algorithm.
     *
     * @param matchingAlgorithm
     * @param stringToLookIn
     *            text to match against
     * @param stringToLookFor
     *            text to match
     * @return
     */
    private boolean doesItMatch(String matchingAlgorithm, String stringToLookIn,
            String stringToLookFor) {
        MatchingAlgorithm m = matchingAlgorithmFactory.getMatchingAlgorithm(matchingAlgorithm);
        if (m == null) {
            throw new UnknownMatchAlgorithm("Matching Algorithm was " + matchingAlgorithm);
        }
        return m.isAMatch(stringToLookIn, stringToLookFor);
    }

    /**
     * Validates the presentation matches the given language. If no language is specified in either
     * the parameter or the presentation, then accept it (more info rather than less).
     *
     * @param languageCode
     * @param p
     * @return
     */
    private boolean isValidLanguage(String languageCode, org.LexGrid.commonTypes.Property p) {
        return LanguageUtil.isMatchLanguage(languageCode, p);
    }

    /**
     * Uses the string name for MatchAlgorithm to get the LexEvs matchAlgorithm name.
     *
     * @param matchAlgorithm
     * @return
     */
    private MatchingAlgorithm getMatchAlgorithm(String matchAlgorithm) {
        MatchingAlgorithm matchingAlgorithm = matchingAlgorithmFactory
                .getMatchingAlgorithm(matchAlgorithm);
        if (matchingAlgorithm == null) {
            throw new UnknownMatchAlgorithm("Matching Algorithm was " + matchingAlgorithm);
        }
        return matchingAlgorithm;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.BrowserOperations#lookupProperties(com.gehc.ce.terminology.cts
     * .dto.ConceptId, java.lang.String[], java.lang.String, java.lang.String, java.lang.String,
     * java.lang.String[])
     */
    public ConceptProperty[] lookupProperties(ConceptId conceptId, String[] properties,
            String matchText, String matchAlgorithmCode, String languageCode, String[] mimeTypes) {
        if (languageCode == null) {
            throw new UnknownLanguageCode("language code was " + languageCode);
        }

        validateMimeTypes(mimeTypes);
        verifyConceptId(conceptId);
        ResolvedConceptReference resolvedConceptReference = queryConcept(conceptId);

        if (resolvedConceptReference == null) {
            throw new UnknownConceptCode(conceptId.getConceptCode());
        }

        List<ConceptProperty> returnPropList = filterProperties(properties, matchText,
                matchAlgorithmCode, languageCode, resolvedConceptReference);

        return returnPropList.toArray(new ConceptProperty[returnPropList.size()]);
    }

    /**
     * @param mimeTypes
     */
    protected void validateMimeTypes(String[] mimeTypes) {
        if (mimeTypes != null) {
            for (int i = 0; i < mimeTypes.length; i++) {
                if (mimeTypes[i] == null || mimeTypes[i] == ""
                        || !mimeTypes[i].contains(this.getSupportedMimeTypes()[i])) {
                    throw new UnknownMimeTypeCode("Invalid mimeType.");
                }
            }
        }
    }

    /**
     * Insure that the conceptId passed in to various services contains values
     *
     * @param conceptId
     *            Code System and concept Code
     *
     * @return
     */
    private void verifyConceptId(ConceptId conceptId) {
        if (conceptId == null) {
            throw new UnknownConceptCode("null");
        }
        if (conceptId.getConceptCode() == null || conceptId.getConceptCode().length() == 0) {
            throw new UnknownConceptCode(conceptId.getConceptCode());
        }
        verifyCodeSystem(conceptId);

    }

    private void verifyCodeSystem(ConceptId conceptId) {
        if (conceptId.getCodeSystemId() == null || conceptId.getCodeSystemId().length() == 0) {
            throw new UnknownCodeSystem(conceptId.getCodeSystemId());
        }
    }

    /**
     * Given various matching parameters, filter all non-matching properties out of those already
     * returned for a concept.
     *
     * @param list
     *            of properties (for the property IDs) to filter the concept's properties. So filter
     *            the concept's list of properties to those with IDs matching the IDs in this list.
     *            If this list is null or empty, then all properties from the concept match this
     *            parameter.
     * @param matchText
     *            text the value of each candidate property should match
     * @param matchAlgorithmCode
     *            the algorithm to use against the matchText
     * @param languageCode
     *            language the property value should be in
     * @param resolvedConceptReference
     * @return
     */
    @SuppressWarnings("unchecked")
	private List<ConceptProperty> filterProperties(String[] properties, String matchText,
            String matchAlgorithmCode, String languageCode,
            ResolvedConceptReference resolvedConceptReference) {

        List<ConceptProperty> returnPropList = new ArrayList<ConceptProperty>();
        
        Concept c = resolvedConceptReference.getReferencedEntry();
        List<Property> propList = new ArrayList();
        
        propList.addAll(safeArrayToList(c.getComment()));
        propList.addAll(safeArrayToList(c.getProperty()));
        propList.addAll(safeArrayToList(c.getDefinition()));
        propList.addAll(safeArrayToList(c.getPresentation()));

        List<String> propCodes = safeArrayToList(properties);

        for (org.LexGrid.commonTypes.Property prop : propList) {
        	Text valueText = prop.getValue();
            String value = valueText != null ? valueText.getContent() : " ";
            boolean match = propCodes.isEmpty() || propCodes.contains(prop.getPropertyName());
            match = match && doesItMatch(matchAlgorithmCode, value, matchText);
            match = match && isValidLanguage(languageCode, prop);
            if (match) {
                ConceptProperty property = ConceptPropertyAdapter.adapt(prop);
                returnPropList.add(property);
            }
        }
        return returnPropList;
    }

    /**
     * Convert an array to a list safely - if the array is null, then return an empty list.
     *
     * @param <T>
     * @param array
     * @return
     */
    private <T> List<T> safeArrayToList(T[] array) {
        if (array == null) {
            return new ArrayList<T>();
        }
        List<T> list = Arrays.asList(array);
        return list;
    }

    /**
     * @param timeLimit
     * @param sizeLimit
     * @return list of code systems
     * @throws TimeoutError
     * @throws UnexpectedError
     */
    @Override
    public CodeSystemIdAndVersions[] getSupportedCodeSystems(long timeLimit, int sizeLimit) {
        CodeSystemIdAndVersions[] codeSystemIdAndVersion = this.getSupportedCodeSystems();
        if (codeSystemIdAndVersion != null && codeSystemIdAndVersion.length <= sizeLimit) {
            return codeSystemIdAndVersion;
        } else if (codeSystemIdAndVersion != null && codeSystemIdAndVersion.length > sizeLimit) {
            List<CodeSystemIdAndVersions> codeSystemList = this.getLimitedSizeList(sizeLimit,
                    codeSystemIdAndVersion);
            return (CodeSystemIdAndVersions[])(codeSystemList
                    .toArray(new CodeSystemIdAndVersions[codeSystemList.size()]));
        } else {

            throw new UnexpectedError();
        }
    }

    /**
     * @param codeSystemId
     *            -
     * @param matchText
     *            -
     * @param matchAlgorithmCode
     *            -
     * @param languageCode
     *            -
     * @param activeConceptsOnly
     *            -
     * @return a list of concept codes that match the supplied search criteria
     * @param timeOut
     * @param sizeLimit
     * @throws UnknownLanguageCode
     * @throws BadlyFormedMatchText
     * @throws UnknownMatchAlgorithm
     * @throws UnknownCodeSystem
     * @throws TimeoutError
     * @throws UnexpectedError
     */
    @Override
    public ConceptId[] lookupConceptCodesByDesignation(String codeSystemId, String matchText,
            String matchAlgorithmCode, String languageCode, boolean activeConceptsOnly,
            long timeOut, int sizeLimit) {
        ConceptId[] conceptId = this.lookupConceptCodesByDesignation(codeSystemId, matchText,
                matchAlgorithmCode, languageCode, activeConceptsOnly);
        if (conceptId != null && conceptId.length <= sizeLimit) {
            return conceptId;
        } else if (conceptId != null && conceptId.length > sizeLimit) {
            List<ConceptId> conceptList = this.getLimitedSizeList(sizeLimit, conceptId);
            return (ConceptId[])conceptList.toArray(new ConceptId[conceptList.size()]);
        } else {

            throw new UnexpectedError();
        }

    }

    /**
     * @param codeSystemId
     *            -
     * @param matchText
     *            -
     * @param matchAlgorithmCode
     *            -
     * @param languageCode
     *            -
     * @param activeConceptsOnly
     *            -
     * @param properties
     *            -
     * @param mimeTypes
     *            -
     * @param timeout
     *            -
     * @param sizeLimit
     *            -
     * @param timeOut
     * @param sizeLimit
     * @return a list of concept codes that match the supplied search criteria
     * @throws UnknownLanguageCode
     * @throws UnknownPropertyCode
     * @throws BadlyFormedMatchText
     * @throws UnknownMatchAlgorithm
     * @throws UnknownMimeTypeCode
     * @throws UnknownCodeSystem
     * @throws TimeoutError
     * @throws UnexpectedError
     */
    @Override
    public ConceptId[] lookupConceptCodesByProperty(String codeSystemId, String matchText,
            String matchAlgorithmCode, String languageCode, boolean activeConceptsOnly,
            String[] properties, String[] mimeTypes, long timeOut, int sizeLimit) {
        // TODO timeOut
        ConceptId[] conceptId = this.lookupConceptCodesByProperty(codeSystemId, matchText,
                matchAlgorithmCode, languageCode, activeConceptsOnly, properties, mimeTypes);
        if (conceptId != null && conceptId.length <= sizeLimit) {
            return conceptId;
        } else if (conceptId != null && conceptId.length > sizeLimit) {
            List<ConceptId> conceptList = this.getLimitedSizeList(sizeLimit, conceptId);
            return (ConceptId[])conceptList.toArray(new ConceptId[conceptList.size()]);
        } else {

            throw new UnexpectedError();
        }

    }

    /**
     * @param expandConceptId
     * @param relationshipCode
     * @param sourceToTarget
     * @param directRelationsOnly
     * @param designationLanguageCode
     * @param timeOut
     * @param sizeLimit
     * @return RelatedCode[]
     * @throws UnknownLanguageCode
     * @throws UnknownRelationshipCode
     * @throws UnknownConceptCode
     * @throws UnknownCodeSystem
     * @throws TimeoutError
     * @throws UnexpectedError
     */
    @Override
    public RelatedCode[] lookupCodeExpansion(ConceptId expandConceptId, String relationshipCode,
            boolean sourceToTarget, boolean directRelationsOnly, String designationLanguageCode,
            long timeOut, int sizeLimit) {
        // TODO timeOut
        RelatedCode[] relatedCode = this.lookupCodeExpansion(expandConceptId, relationshipCode,
                sourceToTarget, directRelationsOnly, designationLanguageCode);
        if (relatedCode != null && relatedCode.length <= sizeLimit) {
            return relatedCode;
        } else if (relatedCode != null && relatedCode.length > sizeLimit) {
            List<RelatedCode> relatedCodeList = this.getLimitedSizeList(sizeLimit, relatedCode);
            return (RelatedCode[])relatedCodeList.toArray(new RelatedCode[relatedCodeList.size()]);
        } else {
            throw new UnexpectedError();
        }

    }

    private <T> List<T> getLimitedSizeList(int size, T... inputArray) {
        if (size <= 0) {
            return new ArrayList<T>();
        }
        return Arrays.asList(inputArray).subList(0, size);
    }
}