/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.impl;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.LexGrid.LexBIG.DataModel.Core.ConceptReference;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.DataModel.InterfaceElements.CodingSchemeRendering;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBInvocationException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.Impl.logging.LgLoggerIF;
import org.LexGrid.LexBIG.Impl.logging.LoggerFactory;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet;
import org.lexgrid.extension.cts.CodeMappingOperations;
import org.lexgrid.extension.cts.lexevs.adapters.AssociationToMappedConceptAdapter;
import org.lexgrid.extension.cts.lexevs.adapters.ConceptIdToConceptReferenceAdapter;
import org.lexgrid.extension.cts.lexevs.repository.LexGridRepository;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationGraph;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationGraphContext;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationMapper;
import org.lexgrid.extension.cts.lexevs.util.template.AssociationRestrictions;
import org.lexgrid.extension.cts.lexevs.util.template.exceptions.LexEvsException;
import org.lexgrid.extension.cts.lexevs.util.template.exceptions.LexEvsExceptionUtil;
import org.lexgrid.extension.cts.terminology.dto.CodeMap;
import org.lexgrid.extension.cts.terminology.dto.ConceptId;
import org.lexgrid.extension.cts.terminology.dto.MappedConceptCode;
import org.lexgrid.extension.cts.terminology.exceptions.AmbiguousMapRequest;
import org.lexgrid.extension.cts.terminology.exceptions.MappingNotAvailable;
import org.lexgrid.extension.cts.terminology.exceptions.UnableToMap;
import org.lexgrid.extension.cts.terminology.exceptions.UnexpectedError;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownCodeSystem;
import org.lexgrid.extension.cts.terminology.exceptions.UnknownConceptCode;
import org.lexgrid.extension.cts.terminology.impl.exceptions.ExceptionHandler;
import org.lexgrid.extension.cts.terminology.impl.exceptions.ExceptionUtil;

/**
 * Adapts the LexBIG/LexEvs API to the CTS {@link CodeMappingOperations} API.
 */
public class LexBigCodeMappingOperations extends IdentifiableServicesImpl implements
        CodeMappingOperations {

    private final LgLoggerIF logger = LoggerFactory.getLogger();

    private String codeMappingRelationsContainerName;

    private LexGridRepository codeMappingRepository;

    private List<ExceptionHandler<Exception>> exceptionHandlers = new ArrayList<ExceptionHandler<Exception>>();

    public List<ExceptionHandler<Exception>> getExceptionHandlers() {
        return exceptionHandlers;
    }

    public void setExceptionHandlers(List<ExceptionHandler<Exception>> exceptionHandlers) {
        this.exceptionHandlers = exceptionHandlers;
    }

    public LexGridRepository getCodeMappingRepository() {
        return codeMappingRepository;
    }

    public void setCodeMappingRepository(LexGridRepository codeMappingRepository) {
        this.codeMappingRepository = codeMappingRepository;
    }

    public String getCodeMappingRelationsContainerName() {
        return codeMappingRelationsContainerName;
    }

    public void setCodeMappingRelationsContainerName(String codeMappingRelationsContainerName) {
        this.codeMappingRelationsContainerName = codeMappingRelationsContainerName;
    }

    /**
     * @see com.gehc.ce.terminology.cts.CodeMappingOperations#getSupportedMaps()
     */
    public CodeMap[] getSupportedMaps() {

        List<CodeMap> codeMaps = new ArrayList<CodeMap>();
        try {
            codeMaps = codeMappingRepository
                    .getCodeMaps(super.lexEVStemplate.getService(), codeMappingRelationsContainerName);
        } catch (Exception e) {
            handle(e);
        }
        return (CodeMap[])codeMaps.toArray(new CodeMap[codeMaps.size()]);
    }

    /**
     * @param e
     */
    private void handle(Exception e) {
        logger.error(e.getMessage(), e);
        ExceptionUtil.handle(e, this.exceptionHandlers);
    }

    /**
     * Returned {@link MappedConceptCode}'s mapQualityCode will be the association name. see
     * {@link CodeMappingOperations#mapConceptCode(ConceptId, String, String)}
     * @param mapName is ignored
     */
    public MappedConceptCode mapConceptCode(ConceptId fromConceptId, String toCodeSystem,
            String mapName) {

        MappedConceptCode mappedConceptCode = null;
        try {
            ConceptReference conceptRef = ConceptIdToConceptReferenceAdapter.adapt(fromConceptId);
            AssociationGraph<MappedConceptCode> mappedCodes = super.lexEVStemplate.queryAssociations(
                    conceptRef, createMapper(), createRestrictions(toCodeSystem));
            mappedConceptCode = findMappedCode(mappedCodes, toCodeSystem, fromConceptId);
        } catch (LexEvsException e) {
            handle(e);
        }

        return mappedConceptCode;
    }

    /**
     * Given the list of mapped codes, find the mapping requested and report any of the defined CTS
     * exceptions.
     *
     * @param mappedCodesGraph
     * @param targetCodeSystem
     * @param sourceCodeSystem
     * @return
     * @throws LBException
     */
    private MappedConceptCode findMappedCode(AssociationGraph<MappedConceptCode> mappedCodesGraph,
            String targetCodeSystem, ConceptId source) {

        List<MappedConceptCode> mappedCodes = mappedCodesGraph.getWhereFocusIsSource();

        if (mappedCodes.size() > 1) {
            throw new AmbiguousMapRequest(gatherPossibleMaps(mappedCodes));
        } else if (mappedCodes.size() < 1) {

            checkValidCode(source);
            checkValidCodeSystem(targetCodeSystem);
            checkMappingAvailable(source, targetCodeSystem);

            // this must occur only if no mappedCodes were found and no other
            // checks explain why
            throw new UnableToMap();
        } else {
            // should only be one then
            return mappedCodes.get(0);
        }
    }

    /**
     * See if we support the given source and target code systems for mapping.
     *
     * @param source
     * @param targetCodeSystem
     * @throws UnexpectedError
     * @throws MappingNotAvailable
     */
    protected void checkMappingAvailable(ConceptId source, String targetCodeSystem) {
        CodeMap[] supportedMaps = getSupportedMaps();
        boolean available = false;
        for (CodeMap codeMap : supportedMaps) {
            if (codeMap.appliesTo(source.getCodeSystemId(), targetCodeSystem)) {
                available = true;
            }
        }
        if (!available) {
            throw new MappingNotAvailable(source.getCodeSystemId(), targetCodeSystem);
        }
    }

    /**
     * See if this concept even exists.
     *
     * @param id
     * @throws LBException
     * @throws Exception
     */
    protected void checkValidCode(ConceptId id) {
        ResolvedConceptReference concept = super.lexEVStemplate.queryConcept(id.getConceptCode(), id
                .getCodeSystemId());
        if (concept == null) {
            throw new UnknownConceptCode(id.getConceptCode());
        }
    }

    /**
     * Validate the given code system is supported by the system.
     *
     * @param sourceCodeSystem
     * @throws LBInvocationException
     */
    protected void checkValidCodeSystem(String sourceCodeSystem) {
        try {
            Iterator<CodingSchemeRendering> iterator = super.lexEVStemplate.getService()
                    .getSupportedCodingSchemes().iterateCodingSchemeRendering();
            while (iterator.hasNext()) {
                if (sourceCodeSystem.equals(iterator.next().getCodingSchemeSummary()
                        .getCodingSchemeURI())) {
                    return;
                }
            }
        } catch (LBException e) {
            LexEvsExceptionUtil.wrapLBException(e);
        }
        throw new UnknownCodeSystem(sourceCodeSystem);
    }

    /**
     *
     * @param mappedCodes
     * @return
     */
    protected String[] gatherPossibleMaps(List<MappedConceptCode> mappedCodes) {
        // TODO: gather the possible maps an inject into exception
        return null;
    }

    /**
     * Restrict the query to mappings to the given target code system.
     *
     * @param targetCodeSystem
     * @return
     */
    protected AssociationRestrictions createRestrictions(final String targetCodeSystem) {
        AssociationRestrictions restrictions = new AssociationRestrictions() {
            public void addRestrictions(CodedNodeGraph nodeGraph) throws LBInvocationException,
                    LBParameterException {
                nodeGraph.restrictToTargetCodeSystem(targetCodeSystem);
            }

            public String getContainer() {
                return getCodeMappingRelationsContainerName();
            }

            public void addRestrictions(CodedNodeSet nodeSet) {}
        };
        return restrictions;
    }

    /**
     * Create an association mapper that will put the association data into a MappedConceptCode.
     *
     * @return
     */
    protected AssociationMapper<MappedConceptCode> createMapper() {
        AssociationMapper<MappedConceptCode> mapper = new AssociationMapper<MappedConceptCode>() {
            @Override
            public MappedConceptCode map(AssociationGraphContext context) {
                return AssociationToMappedConceptAdapter.adapt(context.getAssociation(), context
                        .getAssociate());
            }
        };
        return mapper;
    }

}