/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.LexGrid.LexBIG.DataModel.Collections.CodingSchemeRenderingList;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Impl.logging.LgLoggerIF;
import org.LexGrid.LexBIG.Impl.logging.LoggerFactory;
import org.LexGrid.codingSchemes.CodingScheme;
import org.LexGrid.concepts.Concept;
import org.LexGrid.concepts.Presentation;
import org.lexgrid.extension.cts.RuntimeOperations;
import org.lexgrid.extension.cts.lexevs.adapters.CodingSchemeListToIdAndVersionsArrayAdapter;
import org.lexgrid.extension.cts.lexevs.adapters.CodingSchemeToCodeSystemInfoAdapter;
import org.lexgrid.extension.cts.terminology.dto.CodeSystemIdAndVersions;
import org.lexgrid.extension.cts.terminology.dto.CodeSystemInfo;
import org.lexgrid.extension.cts.terminology.dto.ConceptId;
import org.lexgrid.extension.cts.terminology.dto.StringAndLanguage;
import org.lexgrid.extension.cts.terminology.exceptions.UnexpectedError;
import org.lexgrid.extension.cts.terminology.impl.exceptions.ExceptionHandler;
import org.lexgrid.extension.cts.terminology.impl.exceptions.ExceptionUtil;

public class LexBigRuntimeOperations extends IdentifiableServicesImpl implements RuntimeOperations {

    private final LgLoggerIF logger = LoggerFactory.getLogger();

    private String defaultCodeSystemId;

    private List<ExceptionHandler<Exception>> exceptionHandlers = new ArrayList<ExceptionHandler<Exception>>();

    public List<ExceptionHandler<Exception>> getExceptionHandlers() {
        return exceptionHandlers;
    }

    public void setExceptionHandlers(List<ExceptionHandler<Exception>> exceptionHandlers) {
        this.exceptionHandlers = exceptionHandlers;
    }

/**
     * Determines if a relationship exists between the sourceCode and the targetCode based on the
     * relationshipCode
     *
     * @param relationQualifiers
     *            is ignored.
     */
public boolean areCodesRelated(String codeSystemId, String sourceCode, String targetCode,
            String relationshipCode, String[] relationQualifiers, boolean directRelationsOnly) {

        Boolean codesRelated = false;
        try {
            codesRelated = super.lexEVStemplate.areCodesRelated(codeSystemId, sourceCode, targetCode,
                    relationshipCode, directRelationsOnly);
        } catch (Exception e) {
            handle(e);
        }

        return codesRelated;
    }

    /*
     * (non-Javadoc)
     *
     * @see com.gehc.ce.terminology.cts.RuntimeOperations#getSupportedCodeSystems()
     */
    public CodeSystemIdAndVersions[] getSupportedCodeSystems() {
        try {
            CodingSchemeRenderingList supportedCodingSchemes = super.lexEVStemplate.getService()
                    .getSupportedCodingSchemes();
            
            //if copyright is required, resolveCopyright has to be called here.
            
            if (supportedCodingSchemes == null) {
                return new CodeSystemIdAndVersions[0];
            }
            return CodingSchemeListToIdAndVersionsArrayAdapter.adapt(super.lexEVStemplate.getService(), supportedCodingSchemes);

        } catch (Exception e) {
            handle(e);
            throw new UnexpectedError(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.RuntimeOperations#isConceptIdValid(com.gehc.ce.terminology.cts
     * .dto.ConceptId, boolean)
     */
    public boolean isConceptIdValid(ConceptId conceptId, boolean activeConceptsOnly) {
        checkId(conceptId);

        ResolvedConceptReference queryConcept = queryForConcept(conceptId);

        return checkValidity(activeConceptsOnly, queryConcept);
    }

    /**
     * Find the concept... translate any exceptions to CTS exceptions
     * @param conceptId
     * @return
     */
    private ResolvedConceptReference queryForConcept(ConceptId conceptId) {
        ResolvedConceptReference queryConcept = null;
        try {
            queryConcept = super.lexEVStemplate.queryConcept(conceptId.getConceptCode(), conceptId
                    .getCodeSystemId());
            return queryConcept;
        } catch (Exception e) {
            handle(e);
        }
        return queryConcept;
    }

    /**
     * check if result from concept query results in a valid concept response
     *
     * @param activeConceptsOnly
     * @param queryConcept
     * @return
     */
    private boolean checkValidity(boolean activeConceptsOnly, ResolvedConceptReference queryConcept) {
        boolean returnValue = false;
        if (queryConcept == null) {
            return false;
        }
        if (activeConceptsOnly && queryConcept.getReferencedEntry().isIsActive()) {
            returnValue = true;
        } else if (!activeConceptsOnly) {
            returnValue = true;
        }

        return returnValue;
    }

    /**
     * Check code system id... use default if no code system is given.
     *
     * @param conceptId
     */
    private void checkId(ConceptId conceptId) {
        String id = conceptId.getCodeSystemId();
        if (id == null || id.length() < 1) {
            conceptId.setCodeSystemId(defaultCodeSystemId);
        }
    }

    /**
     * Retrieves detailed information for a CodeSystem/CodingScheme.
     *
     * @param codeSystemId
     * @param codeSystemName
     *
     * @return CodeSystemInfo
     */
    public CodeSystemInfo lookupCodeSystemInfo(String codeSystemId, String codeSystemName) {

        CodeSystemInfo info = new CodeSystemInfo();
        try {
            CodingScheme codingScheme = super.lexEVStemplate.getService().resolveCodingScheme(
                    codeSystemId, null);
            info = CodingSchemeToCodeSystemInfoAdapter.adapt(codingScheme);

        } catch (Exception e) {
            handle(e);
        }
        return info;
    }

    /**
     * The LexBIG Call getPreferredDisplayName() is available, but does not currenlty take in
     * language. However, the call could be modified to provide language (currently uses default
     * language - default language of coding scheme).
     *
     * @return The preferred designation and language. Null if no preferred designation is specified
     *         for this concept
     */
    public StringAndLanguage lookupDesignation(ConceptId conceptId, String ctsLanguageCode) {
        StringAndLanguage returnValue = null;

        String propLanguage = null;
        ResolvedConceptReference resolvedConceptReference = queryForConcept(conceptId);
        Concept actualConcept = resolvedConceptReference.getReferencedEntry();
        Presentation[] presentations = actualConcept.getPresentation();
        for (Presentation p : presentations) {
            /*if (p.isIsPreferred() != null && p.isIsPreferred()
                    && isLanguageOk(ctsLanguageCode, p.getLanguage())) {
                returnValue = new StringAndLanguage(p.getValue().getContent(), p.getLanguage());
            }*/
        	
        	if (p.isIsPreferred() != null && p.isIsPreferred()) {
        		
        		propLanguage = p.getLanguage();
				
        		if (propLanguage == null || "".equals(propLanguage.trim())) {
					try {
						propLanguage = super.lexEVStemplate.getService().resolveCodingScheme(
								resolvedConceptReference.getCodingSchemeName(),
								null).getDefaultLanguage();
					} catch (LBException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
				}
				
				if( isLanguageOk(ctsLanguageCode, propLanguage) ) {
					returnValue = new StringAndLanguage(p.getValue().getContent(), propLanguage);
				}
			}
        }
        return returnValue;

    }

    /**
     * Do the two language codes match?
     *
     * @param ctsLanguage
     * @param lexBigLanguage
     * @return
     */
    private boolean isLanguageOk(String ctsLanguage, String lexBigLanguage) {
        if (ctsLanguage == null) {
            return true;
        } else if (ctsLanguage != null && lexBigLanguage != null
                && lexBigLanguage.equalsIgnoreCase(ctsLanguage)) {
            return true;
        }
        return false;

    }

    /*
     * (non-Javadoc)
     *
     * @see com.gehc.ce.terminology.cts.RuntimeOperations#getSupportedCodeSystems(long, int)
     */
    @Override
    public CodeSystemIdAndVersions[] getSupportedCodeSystems(long timeOut, int sizeLimit) {
        CodeSystemIdAndVersions[] codeSystemIdAndVersion = this.getSupportedCodeSystems();
        if (codeSystemIdAndVersion != null && codeSystemIdAndVersion.length <= sizeLimit) {
            return codeSystemIdAndVersion;
        } else {
            List<CodeSystemIdAndVersions> codeSystemList = Arrays.asList(codeSystemIdAndVersion)
                    .subList(0, sizeLimit - 1);
            return (CodeSystemIdAndVersions[])(codeSystemList
                    .toArray(new CodeSystemIdAndVersions[codeSystemList.size()]));
        }

    }

    /**
     * Handle Exceptions
     *
     * @param e
     */
    private void handle(Exception e) {
        logger.error(e.getMessage(), e);
        ExceptionUtil.handle(e, this.exceptionHandlers);
    }

}