/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.lexevs.repository;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.Impl.dataAccess.ResourceManager;
import org.LexGrid.LexBIG.Impl.dataAccess.SQLInterface;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.codingSchemes.CodingScheme;
import org.LexGrid.naming.Mappings;
import org.LexGrid.naming.SupportedCodingScheme;
import org.LexGrid.naming.SupportedNamespace;
import org.LexGrid.util.sql.lgTables.SQLTableConstants;
import org.lexgrid.extension.cts.terminology.dto.CodeMap;

public class LexGridRepositoryImpl implements LexGridRepository {

	/**
	 * Queries the LexBig Database to find the mappings.
	 * 
	 * @param mappingStr
	 * @return the supportedMimeTypes
	 */
	public List<CodeMap> getCodeMaps(LexBIGService service, String mappingStr) {
		PreparedStatement preparedStatement = null;
		ResultSet resultSet = null;

		List<CodeMap> list = new ArrayList<CodeMap>();

		SQLInterface[] allSQLInterfaces = ResourceManager.instance()
				.getAllSQLInterfaces();
		for (int i = 0; i < allSQLInterfaces.length; i++) {
			try {
				preparedStatement = allSQLInterfaces[i]
						.checkOutPreparedStatement("SELECT DISTINCT "
								+ SQLTableConstants.TBLCOL_CODINGSCHEMENAME
								+ " , "
								+ SQLTableConstants.TBLCOL_SOURCEENTITYCODENAMESPACE
								+ " , "
								+ SQLTableConstants.TBLCOL_TARGETENTITYCODENAMESPACE
								+ " FROM "
								+ allSQLInterfaces[i]
										.getTableName(SQLTableConstants.ENTITY_ASSOCIATION_TO_ENTITY)
								+ " WHERE "
								+ SQLTableConstants.TBLCOL_CONTAINERNAME
								+ " = ?");
				preparedStatement.setString(1, mappingStr);
				resultSet = preparedStatement.executeQuery();
				extractToList(service, resultSet, list);
				resultSet.close();
				resultSet = null;
			} catch (SQLException e) {
				throw new RuntimeException(e);
			} catch (LBException e) {
				throw new RuntimeException(e);
			} finally {
				allSQLInterfaces[i].checkInPreparedStatement(preparedStatement);
			}
		}
		return list;
	}

	/**
	 * @param resultSet
	 * @param list
	 * @throws SQLException
	 * @throws LBParameterException
	 */
	private void extractToList(LexBIGService service, ResultSet resultSet,
			List<CodeMap> list) throws SQLException, LBException {

		CodeMap codeMap = null;
		while (resultSet.next()) {
			codeMap = new CodeMap();

			String csName = resultSet.getString(SQLTableConstants.TBLCOL_CODINGSCHEMENAME);
			String source = resultSet.getString(SQLTableConstants.TBLCOL_SOURCEENTITYCODENAMESPACE);
			String target = resultSet.getString(SQLTableConstants.TBLCOL_TARGETENTITYCODENAMESPACE);

			String sourceEntityURN = null;
			boolean sourcePartOfsuppNameSpace = false;

			String targetEntityURN = null;
			boolean targetPartOfsuppNameSpace = false;

			CodingScheme cs = null;

			try {
				sourceEntityURN = ResourceManager.instance()
						.getURNForExternalCodingSchemeName(source);

			} catch (LBParameterException e) {
				// resolve codingScheme and get supported attributes.
				cs = service.resolveCodingScheme(csName, null);

				if (cs != null) {
					Mappings mappings = cs.getMappings();
					SupportedNamespace[] namespaces = mappings != null ? mappings
							.getSupportedNamespace()
							: new SupportedNamespace[0];
					// check if sourceEntityNamespace is in supportedNamespace.
					// if yes, get the equivalendCodingscheme and lookup into
					// supportedCodingscheme. If present, assign the uri of
					// supportedCodingscheme as sourceEntityURN

					for (int i = 0; i < namespaces.length
							&& !sourcePartOfsuppNameSpace; i++) {
						String namespaceId = namespaces[i].getLocalId();

						if (source.equals(namespaceId)) {

							sourcePartOfsuppNameSpace = true;
							String eqCodingScheme = namespaces[i]
									.getEquivalentCodingScheme();
							source = eqCodingScheme;

							if (eqCodingScheme != null && !eqCodingScheme.trim().equals("")) {
								SupportedCodingScheme[] supCodingSchemes = mappings
										.getSupportedCodingScheme();

								for (int j = 0; j < supCodingSchemes.length; j++) {
									String csLocalId = supCodingSchemes[j]
											.getLocalId();
									if (eqCodingScheme.equals(csLocalId)) {
										sourceEntityURN = supCodingSchemes[j]
												.getUri();
										break;
									}
								}
							}
						}
					}
				}
			}

			codeMap.setFromCodeSystemName(sourceEntityURN);

			if (sourceEntityURN == null) {
				codeMap.setFromCodeSystemVersion(null);
			} else if (!sourcePartOfsuppNameSpace) {
				codeMap.setFromCodeSystemVersion(ResourceManager.instance()
						.getInternalVersionStringFor(source, null));
			}

			try {
				targetEntityURN = ResourceManager.instance()
						.getURNForExternalCodingSchemeName(target);
			} catch (LBParameterException e) {

				if (cs == null) {
					cs = service.resolveCodingScheme(csName, null);
				}

				if (cs != null) {
					Mappings mappings = cs.getMappings();
					SupportedNamespace[] namespaces = mappings != null ? mappings
							.getSupportedNamespace()
							: new SupportedNamespace[0];
					// check if sourceEntityNamespace is in supportedNamespace.
					// if yes, get the equivalendCodingscheme and lookup into
					// supportedCodingscheme. If present, assign the uri of
					// supportedCodingscheme as sourceEntityURN

					for (int i = 0; i < namespaces.length
							&& !targetPartOfsuppNameSpace; i++) {
						String namespaceId = namespaces[i].getLocalId();

						if (target.equals(namespaceId)) {

							targetPartOfsuppNameSpace = true;
							String eqCodingScheme = namespaces[i]
									.getEquivalentCodingScheme();
							target = eqCodingScheme;

							if (eqCodingScheme != null) {
								SupportedCodingScheme[] supCodingSchemes = mappings
										.getSupportedCodingScheme();

								for (int j = 0; j < supCodingSchemes.length; j++) {
									String csLocalId = supCodingSchemes[j]
											.getLocalId();
									if (eqCodingScheme.equals(csLocalId)) {
										targetEntityURN = supCodingSchemes[j]
												.getUri();
										break;
									}
								}
							}
						}
					}
				}
			}

			codeMap.setToCodeSystemName(targetEntityURN);

			if (targetEntityURN == null) {
				codeMap.setToCodeSystemVersion(null);
			} else if (!targetPartOfsuppNameSpace) {
				codeMap.setToCodeSystemVersion(ResourceManager.instance()
						.getInternalVersionStringFor(target, null));
			}
			
			codeMap.setMapName(source + "-" + target);
			codeMap.setMapDescription(source + "-" + target);
			codeMap.setFromCodeSystemId(source);
			codeMap.setToCodeSystemId(target);

			list.add(codeMap);
		}
	}
}