/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.lexevs.util.template;

import java.util.Iterator;
import java.util.List;

import org.LexGrid.LexBIG.DataModel.Collections.ResolvedConceptReferenceList;
import org.LexGrid.LexBIG.DataModel.Core.ConceptReference;
import org.LexGrid.LexBIG.DataModel.Core.ResolvedConceptReference;
import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeGraph;
import org.LexGrid.LexBIG.LexBIGService.CodedNodeSet;
import org.LexGrid.LexBIG.LexBIGService.LexBIGService;
import org.LexGrid.LexBIG.Utility.Constructors;
import org.lexgrid.extension.cts.lexevs.util.template.exceptions.LBCheckedExceptionBlock;
import org.lexgrid.extension.cts.lexevs.util.template.exceptions.LexEvsExceptionUtil;

public class LexEvsTemplate implements LexEvsOperations {

    private LexBIGService service;

    private int maxResults = ResolveParams.DEFAULT_MAX_RESULTS;

    private GraphExtractor graphExtractor = new GraphExtractorImpl();

    private NodeResolver resolver = new NodeResolverImpl();

    private LexEvsProxyFactory proxyFactory = new DynamicLexEvsProxyFactory();

    private String defaultAssociationContainer;

/**
     * The default Restrictions in case none is supplied. Nothing is restricted in this. And the
     * Container, in case of graph queries, is the default container.
     */
private AssociationRestrictions defaultRestrictions = new AssociationRestrictions() {
        public void addRestrictions(CodedNodeGraph nodeSet) {}

        public String getContainer() {
            return defaultAssociationContainer;
        }

        public void addRestrictions(CodedNodeSet nodeSet) {}
    };

    private ConceptExtractor conceptExtractor = new ConceptExtractorImpl();

    /**
     * @return the conceptExtractor
     */
    public ConceptExtractor getConceptExtractor() {
        return conceptExtractor;
    }

    /**
     * @param conceptExtractor
     *            the conceptExtractor to set
     */
    public void setConceptExtractor(ConceptExtractor conceptExtractor) {
        this.conceptExtractor = conceptExtractor;
    }

    public GraphExtractor getGraphExtractor() {
        return graphExtractor;
    }

    public void setGraphExtractor(GraphExtractor graphExtractor) {
        this.graphExtractor = graphExtractor;
    }

    public NodeResolver getResolver() {
        return resolver;
    }

    public void setResolver(NodeResolver resolver) {
        this.resolver = resolver;
    }

    public String getDefaultAssociationContainer() {
        return defaultAssociationContainer;
    }

    public void setDefaultAssociationContainer(String defaultAssociationContainer) {
        this.defaultAssociationContainer = defaultAssociationContainer;
    }

    /*
     * (non-Javadoc)
     *
     * @see com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#getService()
     */
    public LexBIGService getService() {
        return this.service;
    }

    /*
     * (non-Javadoc)
     *
     * @see com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#getMaxResults()
     */
    public int getMaxResults() {
        return maxResults;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#setMaxResults(int)
     */
    public void setMaxResults(int maxResults) {
        this.maxResults = maxResults;
    }

    public LexEvsProxyFactory getProxyFactory() {
        return proxyFactory;
    }

    public void setProxyFactory(LexEvsProxyFactory proxyFactory) {
        this.proxyFactory = proxyFactory;
    }

    /**
     * Constructor: the LexBIGService is certainly required in all cases as this is really a wrapper
     * of the service.
     *
     * @param service
     */
    public LexEvsTemplate(LexBIGService service) {
        this.service = service;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#queryConceptList(java
     * .lang.String, com.gehc.ce.terminology.cts.impl.lexevs.util.template.ConceptMapper,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.Restrictions,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.ResolveParams)
     */
    public <T> List<T> queryConceptList(String codeSystem, ConceptMapper<T> mapper,
            Restrictions restrictions, ResolveParams resolveParams) {

        CodedNodeSet nodeSet = createNodeSet(codeSystem, restrictions);
        ResolvedConceptReferenceList concepts = resolveToList(nodeSet, resolveParams);
        return conceptExtractor.extract(mapper, concepts);
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#queryConceptList(java
     * .lang.String, com.gehc.ce.terminology.cts.impl.lexevs.util.template.ConceptMapper,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.Restrictions)
     */
    public <T> List<T> queryConceptList(String codeSystem, ConceptMapper<T> mapper,
            Restrictions restrictions) {
        return queryConceptList(codeSystem, mapper, restrictions, new ResolveParams(maxResults));
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#queryAssociations(
     * org.LexGrid.LexBIG.DataModel.Core.ConceptReference,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.AssociationMapper,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.AssociationRestrictions)
     */
    public <T> AssociationGraph<T> queryAssociations(ConceptReference concept,
            AssociationMapper<T> mapper, AssociationRestrictions restrictions) {
        return queryAssociations(concept, mapper, restrictions, new ResolveParams(maxResults));
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#queryAssociations(
     * org.LexGrid.LexBIG.DataModel.Core.ConceptReference,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.AssociationMapper,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.AssociationRestrictions,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.ResolveParams)
     */
    public <T> AssociationGraph<T> queryAssociations(ConceptReference concept,
            AssociationMapper<T> mapper, AssociationRestrictions restrictions,
            ResolveParams resolveParams) {
        CodedNodeGraph nodeGraph = createNodeGraph(concept, restrictions);
        ResolvedConceptReferenceList relatedConcepts = resolveGraphAsList(nodeGraph, concept,
                resolveParams);
        return createGraph(mapper, resolveParams, relatedConcepts);
    }

    /**
     * Execute extraction of the LexEVS graph and populate the {@link AssociationGraph} with
     * results.
     *
     * @param <T>
     * @param mapper
     * @param resolveParams
     * @param relatedConcepts
     * @return
     */
    protected <T> AssociationGraph<T> createGraph(AssociationMapper<T> mapper,
            ResolveParams resolveParams, ResolvedConceptReferenceList relatedConcepts) {
        AssociationGraph<T> graph = new AssociationGraph<T>();
        if (resolveParams.isResolveGraphBackward()) {
            graph.setWhereFocusIsTarget(graphExtractor.extractBackward(relatedConcepts, mapper));
        }
        if (resolveParams.isResolveGraphForward()) {
            graph.setWhereFocusIsSource(graphExtractor.extractForward(relatedConcepts, mapper));
        }
        return graph;
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#queryConcept(java.
     * lang.String, java.lang.String)
     */
    public ResolvedConceptReference queryConcept(String code, String codeSystem) {
        return queryConcept(code, codeSystem, new ResolveParams(maxResults));
    }

    /*
     * (non-Javadoc)
     *
     * @see
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#queryConcept(java.
     * lang.String, java.lang.String,
     * com.gehc.ce.terminology.cts.impl.lexevs.util.template.ResolveParams)
     */
    public ResolvedConceptReference queryConcept(String code, String codeSystem,
            ResolveParams resolveParams) {
        Restrictions restrictions = new CodeRestriction(code, codeSystem);
        CodedNodeSet nodeSet = createNodeSet(codeSystem, restrictions);
        ResolvedConceptReferenceList list = resolveToList(nodeSet, resolveParams);
        Iterator<ResolvedConceptReference> iterator = list.iterateResolvedConceptReference();
        if (iterator.hasNext()) {
            return iterator.next();
        }
        return null;
    }

    /**
     * Execute a simple resolve as list on a CodedNodeGraph. Conveniently simplifies your code.
     *
     * @param nodeGraph
     * @param focus
     * @return
     */
    protected ResolvedConceptReferenceList resolveGraphAsList(final CodedNodeGraph nodeGraph,
            final ConceptReference focus, ResolveParams resolveParams) {
        final ResolveParams params = ensureResolveParams(resolveParams);
        return LexEvsExceptionUtil
                .wrapLBException(new LBCheckedExceptionBlock<ResolvedConceptReferenceList>() {
                    public ResolvedConceptReferenceList executeBlock() throws LBException {
                        return resolver.resolveGraphAsList(nodeGraph, focus, params);
                    }
                });
    }

    /**
     * Simple resolve to list for a CodedNodeSet.
     *
     * @param nodeSet
     * @return
     */
    protected ResolvedConceptReferenceList resolveToList(final CodedNodeSet nodeSet,
            ResolveParams resolveParams) {
        final ResolveParams params = ensureResolveParams(resolveParams);
        return LexEvsExceptionUtil
                .wrapLBException(new LBCheckedExceptionBlock<ResolvedConceptReferenceList>() {
                    public ResolvedConceptReferenceList executeBlock() throws LBException {
                        return resolver.resolveToList(nodeSet, params);
                    }
                });
    }

    /**
     * Make sure default resolve parameters are set if not given
     *
     * @param resolveParams
     * @return
     */
    private ResolveParams ensureResolveParams(ResolveParams resolveParams) {
        if (resolveParams == null) {
            return new ResolveParams(maxResults);
        } else {
            return resolveParams;
        }
    }

    /**
     * Creates a CodeNodeGraph complete with restrictions applied. Ready to resolve.
     *
     * @param concept
     * @param restrictions
     * @return
     * @throws LBException
     */
    protected CodedNodeGraph createNodeGraph(final ConceptReference concept,
            AssociationRestrictions restrictions) {

        if (restrictions == null) {
            restrictions = defaultRestrictions;
        }

        final AssociationRestrictions fr = restrictions;
        CodedNodeGraph nodeGraph = LexEvsExceptionUtil.wrapLBException(createNodeGraphBlock(
                concept, fr));
        return nodeGraph;
    }

    /**
     * Create an executable {@link LBCheckedExceptionBlock} to create the node graph so any
     * exceptions are wrapped.
     *
     * @param concept
     * @param fr
     * @return
     */
    private LBCheckedExceptionBlock<CodedNodeGraph> createNodeGraphBlock(
            final ConceptReference concept, final AssociationRestrictions fr) {
        return new LBCheckedExceptionBlock<CodedNodeGraph>() {
            public CodedNodeGraph executeBlock() throws LBException {
                CodedNodeGraph graph = service.getNodeGraph(concept.getCodingSchemeName(), null, fr
                        .getContainer());
                CodedNodeGraph proxy = proxyFactory.getResolveSuppressingProxy(graph);
                fr.addRestrictions(proxy);
                return graph;
            }
        };
    }

    /**
     * Creates a CodedNodeSet complete with restrictions applied. Ready to resolve.
     *
     * @param codeSystem
     * @param restrictions
     * @return
     * @throws LBException
     */
    public CodedNodeSet createNodeSet(final String codeSystem, Restrictions restrictions) {

        if (restrictions == null) {
            restrictions = defaultRestrictions;
        }

        CodedNodeSet set = LexEvsExceptionUtil.wrapLBException(createNodeSetBlock(codeSystem,
                restrictions));

        return set;
    }

    /**
     * create node set in a {@link LBCheckedExceptionBlock} to wrap any exceptions
     * @param codeSystem
     * @param restrictions
     * @return
     */
    private LBCheckedExceptionBlock<CodedNodeSet> createNodeSetBlock(final String codeSystem,
            final Restrictions restrictions) {
        return new LBCheckedExceptionBlock<CodedNodeSet>() {

            public CodedNodeSet executeBlock() throws LBException {
                CodedNodeSet set = service.getCodingSchemeConcepts(codeSystem, null);
                CodedNodeSet proxy = proxyFactory.getResolveSuppressingProxy(set);
                restrictions.addRestrictions(proxy);
                return set;
            }
        };
    }

    /* (non-Javadoc)
     * @see com.gehc.ce.terminology.cts.impl.lexevs.util.template.LexEvsOperations#areCodesRelated(java.lang.String,
     * java.lang.String, java.lang.String, java.lang.String, java.lang.String[], boolean)
     */
    @Override
    public Boolean areCodesRelated(String codeSystemId, String conceptCode, String targetCode,
            String relationshipCode, boolean directRelationsOnly) {
        boolean areRelated = false;
        try {
            CodedNodeGraph nodeGraph = createNodeGraph(Constructors.createConceptReference(conceptCode, codeSystemId), null);
            areRelated = nodeGraph.areCodesRelated(Constructors.createNameAndValue(relationshipCode, null),
                    Constructors.createConceptReference(conceptCode, codeSystemId),
                    Constructors.createConceptReference(targetCode, codeSystemId), directRelationsOnly);

        } catch (LBException e) {
            LexEvsExceptionUtil.wrapLBException(e);
        }
        return areRelated;
    }
}