/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.lexevs.util.template.exceptions;

import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Map;

import org.LexGrid.LexBIG.Exceptions.LBException;
import org.LexGrid.LexBIG.Exceptions.LBInvocationException;
import org.LexGrid.LexBIG.Exceptions.LBParameterException;
import org.LexGrid.LexBIG.Exceptions.LBResourceUnavailableException;
import org.LexGrid.LexBIG.Exceptions.LBUnsupportedOperationException;

/**
 * utility for simplifying exception handling code
 */
public final class LexEvsExceptionUtil {

    private static Map<String, Class< ? extends LexEvsException>> exceptionTranslation = new HashMap<String, Class< ? extends LexEvsException>>();

    static {
        exceptionTranslation.put(LBParameterException.class.getName(),
                LexEvsParameterException.class);
        exceptionTranslation.put(LBInvocationException.class.getName(),
                LexEvsInvocationException.class);
        exceptionTranslation.put(LBResourceUnavailableException.class.getName(),
                LexEvsResourceUnavailableException.class);
        exceptionTranslation.put(LBUnsupportedOperationException.class.getName(),
                LexEvsUnsupportedOperationException.class);
    }

    private LexEvsExceptionUtil() {}

    /**
     * Catch all LBExceptions and wrap them in a runtime exception, decorating them with information
     * if possible. Akin to Spring JDBC Exceptions.
     *
     * @param <T>
     *            type to return from the given code block.
     * @param block
     *            the block of java code to execute that throws these exceptions.
     * @return the return value of the block.
     */
    public static <T> T wrapLBException(LBCheckedExceptionBlock<T> block) {
        try {
            return block.executeBlock();
        } catch (LBException lbe) {
            wrapLBException(lbe);
        }
        return null;
    }

    /**
     * Wrap the given {@link LBException} with one of the template's runtime exceptions.
     *
     * @param lbe
     */
    public static void wrapLBException(LBException lbe) {
        Class< ? extends LexEvsException> eClazz = exceptionTranslation.get(lbe.getClass()
                .getName());
        if (eClazz != null) {
            Object newInstance = instantiateRuntimeException(lbe, eClazz);

            if (newInstance != null && newInstance instanceof LexEvsException) {
                throw (LexEvsException)newInstance;
            }
        }

        throw new LexEvsException(lbe);
    }

    /**
     * Instantiate the runtime exception class to wrap the checked exception
     *
     * @param lbe
     * @param eClazz
     * @return
     */
    private static Object instantiateRuntimeException(LBException lbe,
            Class< ? extends LexEvsException> eClazz) {
        Object newInstance = null;
        try {
            Constructor< ? > constructor = eClazz.getConstructor(lbe.getClass());
            newInstance = constructor.newInstance(lbe);
        } catch (Exception e1) {
            // TODO
            e1.printStackTrace();
        }
        return newInstance;
    }
}