/*
 * Copyright: (c) 2004-2009 Mayo Foundation for Medical Education and 
 * Research (MFMER). All rights reserved. MAYO, MAYO CLINIC, and the
 * triple-shield Mayo logo are trademarks and service marks of MFMER.
 *
 * Except as contained in the copyright notice above, or as used to identify 
 * MFMER as the author of this software, the trade names, trademarks, service
 * marks, or product names of the copyright holder shall not be used in
 * advertising, promotion or otherwise in connection with this software without
 * prior written authorization of the copyright holder.
 * 
 * Licensed under the Eclipse Public License, Version 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at 
 * 
 * 		http://www.eclipse.org/legal/epl-v10.html
 * 
 */
package org.lexgrid.extension.cts.terminology.util;

import java.beans.PropertyEditorSupport;

import org.apache.log4j.Logger;
import org.lexgrid.extension.cts.terminology.dto.CTSVersionId;

/**
 * A javabeans property editor for a CTSVersionId object. This is primarily used for allowing easier
 * Spring configuration. For a property typed as a CTSVersionId, the spring xml property config can
 * just use a string value of 1.2 or something and this class will convert that to a CTSVersionId
 * object.
 */
public class CTSVersionIdPropertyEditor extends PropertyEditorSupport {

    private Logger log = Logger.getLogger(this.getClass());

    /**
     * Uses a period (".") as a delimeter in a string value to break the version number into major
     * and minor components.
     * @param text
     * @exception throws IllegalArgumentException
     *
     */
    @Override
    public void setAsText(String text) {

        if (text == null || text.length() < 1) {
            log.warn("No CTSVersionId given, using 0.0");
            setValue(createVersionId("0", "0"));
            return;
        }

        String[] components = text.split("\\.");

        if (components.length < 2) {
            setValue(createVersionId(components[0], "0")); // assume 0
            return;
        }
        if (components.length > 2) {
            log.warn("A CTSVersionId only uses two version components,"
                    + " such as 1.2 - not 1.2.0.4.  Truncating version number.");
        }
        setValue(createVersionId(components[0], components[1]));
    }

    /**
     * Creates a CTSVersionId object given the two version number components. If either is
     * non-numeric, 0 is substituted.
     *
     * @param major
     * @param minor
     * @return
     */
    public CTSVersionId createVersionId(String major, String minor) {

        CTSVersionId id = new CTSVersionId();
        try {
            id.setMajor(Short.valueOf(major));
        } catch (NumberFormatException e) {
            id.setMajor((short)0);
        }
        try {
            id.setMinor(Short.valueOf(minor));
        } catch (NumberFormatException e) {
            id.setMinor((short)0);
        }
        return id;
    }

    @Override
    /**
     * Reassembles the textual representation, inserting the period (".") to delimit the version
     * number components.
     */
    public String getAsText() {
        CTSVersionId id = (CTSVersionId)getValue();
        return id.getMajor() + "." + id.getMinor();
    }
}